/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: gradient-flow-cf.c
*** Author: John Y. A. Wang
*** Date: 1991
*** Copyright 1991, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description:

***   Coarse to fine local motion estimation using pyramid
***   hierarchial iteration algothrim. Estimation is a
***   gradient flow alogorithm described by Lucas and Kanade
***   user passes temporary memory to function, if memory size
***   less than 10*imagesize then more memory is allocated.
***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: grad-flow-cf.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1991, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <memory.h>
#include "pyramid.h"
#include "motion.h"

/************************************************************************/
/* im1,im2:       input image
** xflow, yflow:  result flow images
** xdim,ydim:     dimensions of images
** level:         level of corase to fine. USE: 2 for small displacements]
** method:        array of floats that indicate number of time to iterate at each
**                level. USE: static float method[3]={1.0,1.0,1.0}
** flowlevel:     indicates how big the integration window. USE:  1.0.
** f:             biasing factor for noise in image: USE 1e-5.
** ttemp:         temporary memory of floats  10x size as image.
** ttempsize:     size of ttemp memory.
** clipmin,clipmax:  min and max of flow estimates. clips at these values.
** xinit, yinit:    initial estimates of flow. float images.
** sum_filt,fdim:   summation filter and dimensions. linear array.
** pyr_filt,pfdim:  pyramid subsampling filter and dimensions. linear array.
** exp_filt,expfdim: pyramid expanding filter and dimensions. linear array.
** thresh = 1:
** stat:       return covariances of estimates in the ttemp if stat=1.
*/
/************************************************************************/

void internal_coarse_fine_gradient_flow
  (im1, im2, xflow, yflow, xdim, ydim, level,method,
   flowlevel,f,ttemp,ttempsize,xbias,ybias,
   clipmin,clipmax,xinit,yinit,
   sum_filt,fdim,pyr_filt,pfdim,exp_filt,expfdim,
   thresh,stat
   /* tempdbg1,tempdbg2 */
   )
  float *im1, *im2;
  register float *xflow, *yflow;
  register int xdim,ydim;
  float *ttemp;
  int ttempsize;
  int level,flowlevel;
  float *method;
  float f;            /* factor */
  register float clipmin, clipmax;
  /*  float *tempdbg1,*tempdbg2; */
  float xbias,ybias;
  float *xinit,*yinit;
  float *sum_filt,*pyr_filt,*exp_filt;
  int fdim,pfdim,expfdim;
  float thresh;
  int stat;
{
/*  static float g_filt[5]={0.0625, 0.25, 0.375, 0.25, 0.0625}; */
/*  static float imp_filt[1]={1.0}; */
  Pyramid pyr1;        /* pyr of 1st image */
  Pyramid pyr2;        /* pyr of 2nd image */
  register float *xtemp,*ytemp;
  register int xsize, ysize,i,j;
  register int im_size = xdim*ydim;
  float *temp;
  int temp_image = 8+2;   /* requires 8+2 temporary images */
  int tempsize=8*im_size;
  int allocate = 0;
  float factor;
  float *ptr1,*ptr2;
  int istat;

/*
  fprintf(stderr,"thresh=%4.2f pyrstat=%d  ",thresh,stat);

  fprintf(stderr,"level=%d,flowlevel=%d,xbias=%6.2f,ybias=%6.2f,xinit=%d,yinit=%d,clipmin=%6.2f,clipmax=%6.2f\n",
	 level,flowlevel,xbias,ybias,xinit,yinit,clipmin,clipmax); 
  fprintf(stderr,"xdim=%d,ydim=%d,f=%8.2e,method=(%d %d %d %d)\n"
	 ,xdim,ydim,f,(int)method[0],(int)method[1],(int)method[2],(int)method[3]);
  fprintf(stderr,"sum_filt: ("); for(i=0;i<fdim;fprintf(stderr,"%6.3f ",sum_filt[i]),i++); fprintf(stderr,")\n");
  fprintf(stderr,"pyr_filt: ("); for(i=0;i<pfdim;fprintf(stderr,"%6.3f ",pyr_filt[i]),i++); fprintf(stderr,")\n");
  fprintf(stderr,"exp_filt: ("); for(i=0;i<expfdim;fprintf(stderr,"%6.3f ",exp_filt[i]),i++); fprintf(stderr,")\n");
  fprintf(stderr,"im1=%d,im2=%d,xflow=%d,yflow=%d,ttemp=%d,ttempsize=%d\n",im1,im2,xflow,yflow,ttemp,ttempsize);
*/
	 
  if(ttempsize < im_size*temp_image) {
    fprintf(stderr,"\ncoarse-fine-gradient-flow: allocate temp image memory");
    fprintf(stderr,"\nsupplied=%d; required=%d\n",
	    ttempsize, im_size*temp_image);

    if((temp = (float *) (malloc(im_size*temp_image*sizeof(float)))) == NULL) {
      fprintf(stderr,"\nmalloc: cannot allocate memory for temporary image\n");
      return;
    }
    allocate = 1;
    fprintf(stderr,"\nallocated %d bytes of memory for temporary image\n",
                im_size*temp_image*sizeof(float)); 
  }
  else
    temp = ttemp;
  
  xtemp=temp+8*im_size;
  ytemp=temp+9*im_size;

  /* make initial estimates based on the level of analyses in the pyarmid.
  ** Generate the x and  y estimates for the starting level and then put 
  ** them in xtemp and ytemp where the intermediate flows are stored
  */

  if(((xinit!=NULL) && (yinit!=NULL))) {
    fprintf(stderr,"\n using init flow\n");
    make_pyramid(xinit,level,xdim,ydim,&pyr1,xtemp,temp,1,
		 pyr_filt,pfdim);
    make_pyramid(yinit,level,xdim,ydim,&pyr2,ytemp,temp,1,
		 pyr_filt,pfdim);
    xsize=pyr1.level[level].xdim;
    ysize=pyr1.level[level].ydim;

    xsize=xsize*ysize;
    ptr1=pyr1.level[level].array;
    ptr2=pyr2.level[level].array;
    for(i=0,factor=1;i<level;i++,factor*=2);
    
    /*    fprintf(stderr,"\ncalculating init flow (factor=%d)",(int)factor); */
    for(i=0;i<xsize;i++) {
      xflow[i]=ptr1[i]/factor;
      yflow[i]=ptr2[i]/factor;
    }
  }
  else {
    for(i=0;i<im_size;i++) {
      xflow[i]=0; yflow[i]=0;
    }
  }
  /* Now make the image pyramids.
     Use third to last temp image for the pyramid level>0. This is safe because temp
     space used by the flow routine will use only (5* 1/4 * im_size) for level>0,
     since images on level 1 are 1/4 size.
  */
  
/*  fprintf(stderr,"\nmaking pyramids.... ");  */

  make_pyramid(im1,level,xdim,ydim,&pyr1,(temp+7*im_size),temp,1,
	       pyr_filt,pfdim);
  make_pyramid(im2,level,xdim,ydim,&pyr2,(temp+(int)(7.5*(float)im_size)),temp,1,
	       pyr_filt,pfdim);

/*  fprintf(stderr,"(%e,%e)",clipmin,clipmax); */
  for(i=level; i>=0; i--) {
    xsize=pyr1.level[i].xdim;
    ysize=pyr1.level[i].ydim;

    for(j=0;j<method[i];j++) {
      if((i==0)&&(j==method[i]-1))
	istat=stat;
      else istat=0;
      bicubic_warp(pyr1.level[i].array,xtemp,yflow,xflow,ysize,xsize);
      gradient_flow_c(xtemp, pyr2.level[i].array,
		      xtemp,ytemp,xsize,ysize,flowlevel,sum_filt,fdim,f,
				       temp,tempsize,xbias,ybias,istat);
      internal_update_flow(xflow,yflow,xtemp,ytemp,xsize,ysize,clipmin,clipmax);
    }
    if(i>0) {
      internal_expand_flow(xflow,yflow,xflow,yflow,temp,xsize,ysize,
			   pyr1.level[i-1].xdim,pyr1.level[i-1].ydim,
			   exp_filt,expfdim,thresh
			   );
    }
  }
  if (allocate)
    free(temp);
}

