/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/***************************************************************
*** File: gradient-flow.c
*** Author: John Y. A. Wang
*** Date: 1991
*** Copyright 1991, Massachusetts Institute of Technology, All Rights Reserved.
***
*** Description:

***   gradient flow alogorithm described by Lucas and Kanade
***   user passes temporary memory to function, if memory size
***   less than 8*imagesize then more memory is allocated

***************************************************************/

static  char *copyright="\
\n\n***********************************************\n\
File: gradient-flow.c\n\
Author: John Y. A. Wang.\n\
The MIT Media Lab.\n\
Copyright (c) 1991, All Rights Reserved.\n\
***********************************************\n\n";

#include <stdio.h>
#include <math.h>
#include <memory.h>
#include "motion.h"

/******************************************************
  COMMENTS:
  --------

  im1, im2:    input images  
  xflow,yflow: output flow images
  xdim, ydim:  images dimensions
  flowlevel:   filter window blur level
  sum_filt:    the summation filter
  f:           singular value bias (estimate noise)
  ttemp:       temporary image for internal operations.
               This should be 8*size_of_input to avoid having to
	       allocate new memory.
  tempsize:    size of temporary image
  xbias, ybias: bias analysis to x or y. These are cost values. Higher
                values bias estimates to zero.
  stat:        == 1 if statistics are to be calculated. Results are returned
               in the temporary memory passed by the user.
               Cov(xflow)       : offset = 0 
               Cov(yflow)       : offset = 1
               Cov(xflow.yflow) : offset = 2
	       1/det(Cov-mat)   : offset = 5
	       Var(noise)       : offset = 7
*********************************************************************/

void gradient_flow_c(im1,im2,xflow,yflow,xdim,ydim,flowlevel,sum_filt,fdim,f,
		ttemp,tempsize,xbias,ybias,stat)
  float *im1, *im2;
  register float *xflow, *yflow;
  float *sum_filt;
  int xdim,ydim,flowlevel,fdim;
  int tempsize;
  float *ttemp;
  float f;
  register float xbias, ybias;
  int stat;
{
  register float denom;
  register int i;

  float *images;
  register float *dx2,*dy2;
  register float *dxdy,*dxdt,*dydt,*dt;
  register float *temp, *temp1;
  int num_temp=8;
  int im_size = xdim*ydim;
  int allocate = 0;

  static float d_filt[2]  = {-1.0, 1.0};
  static float s_filt[2]  = {0.5,  0.5};
  static float s2_filt[2] = {1.0,  1.0};
  char *func="gradient_flow_c()";

/*  fprintf(stderr,"stat=%d ",stat); */
  if(tempsize < im_size*num_temp) {
    if(stat==1) {
      fprintf(stderr,"%s: Please supply ttemp to get statistics on analysis.\n",
	      func); return; }
    fprintf(stderr,"\n%s: allocating ttemp memory \n",func);
    fprintf(stderr,"\nsupplied=%d; required=%d\n",tempsize, im_size*num_temp);
    if((images = (float *) (malloc(num_temp*im_size*sizeof(float)))) == NULL) {
      fprintf(stderr,"\n%s: cannot allocate ttemp memory\n",func);
      return;
    }
    allocate = 1;
  }
  else
    images = ttemp;
/*
  fprintf(stderr,"(%d,%d) %d, %e (%e,%e)\n",
	 xdim, ydim, flowlevel,f,xbias,ybias);
  fprintf(stderr,"%d,%d,%d,%d\n",im1, im2, xflow, yflow,ttemp); 
*/
  dx2  = images;
  dy2  = images + 1*im_size;
  dxdy = images + 2*im_size;
  dxdt = images + 3*im_size;
  dydt = images + 4*im_size;
  dt   = images + 5*im_size;
  temp = images + 6*im_size;
  temp1 = images + 7*im_size;

  if(f<0) f=-1*f;  /* get abs. val. of f */

  add_im(im1,im2,temp1,im_size);
  /* x-derivative */
  apply_sep_filter_reflect(temp1,dx2,temp,xdim,ydim,d_filt,s_filt,2,2,1,1); 

  /* y-derivative */
  apply_sep_filter_reflect(temp1,dy2,temp,xdim,ydim,s_filt,d_filt,2,2,1,1); 

  /* t-derivative */
  sub_im(im1,im2,dt,im_size);
  apply_sep_filter_reflect(dt,dt,temp,xdim,ydim,s2_filt,s_filt,2,2,1,1);
  
  mul_im(dx2,dt, dxdt,im_size);
  blur_image(dxdt,dxdt,temp,xdim,ydim,flowlevel,sum_filt,fdim);
  mul_im(dy2,dt, dydt,im_size); 
  blur_image(dydt,dydt,temp,xdim,ydim,flowlevel,sum_filt,fdim);
  mul_im(dx2,dy2,dxdy,im_size);
  blur_image(dxdy,dxdy,temp,xdim,ydim,flowlevel,sum_filt,fdim);
  mul_im(dx2,dx2,dx2,im_size);
  blur_image(dx2,  dx2,temp,xdim,ydim,flowlevel,sum_filt,fdim);
  mul_im(dy2,dy2,dy2,im_size);
  blur_image(dy2,  dy2,temp,xdim,ydim,flowlevel,sum_filt,fdim);

/*  fprintf(stderr,"calculate-flow\n");*/

  if(stat == 1) {   /* save input images for calculations later */
    copy_im(im1,temp,im_size);
    copy_im(im2,temp1,im_size);
  }

  for (i=0;i<im_size;i++) {
    denom=(dx2[i]+xbias)*(dy2[i]+ybias)-dxdy[i]*dxdy[i];
    dt[i] = denom;               /* save determinant */
    if (denom > 0) denom += f;   /* biasing factor in denom */
    else denom -= f;
    xflow[i] = (dxdt[i]*(dy2[i]+ybias) - dydt[i]*dxdy[i])/denom;
    yflow[i] = (dydt[i]*(dx2[i]+xbias) - dxdt[i]*dxdy[i])/denom;
  }
  if(stat == 1) {
    bicubic_warp(temp,dxdt,yflow,xflow,ydim,xdim);
    sub_im(dxdt,temp1,temp1,im_size);
    mul_im(temp1,temp1,temp1,im_size);
    blur_image(temp1,temp1,temp,xdim,ydim,flowlevel,sum_filt,fdim); /* noise-estimate */   
    div_im(dx2,dt,temp,im_size); div_im(dy2,dt,dx2,im_size);  
    mul_im(temp1,temp,dy2,im_size);  /* Cy */
    mul_im(temp1,dx2,dx2,im_size); /* Cx */
    div_im(dxdy,dt,dxdy,im_size); mul_im(temp1,dxdy,dxdy,im_size); /* Cxy */
  }
  
 if(allocate)
   free(images); 
}

