/******************************************************************

"Copyright  1994 by the Massachusetts  Institute  of  Technology.  All
rights reserved.

Developed by John Y. A. Wang at the  Media Laboratory, MIT, Cambridge,
Massachusetts, with  support  from Secom Ltd., Goldstar Ltd.,  and the
Television of Tomorrow Program.

For use by authorized users only.

This distribution is approved  by Nicholas Negroponte, Director of the
Media Laboratory, MIT.

Permission to use, copy, or modify this software and its documentation
for educational and research purposes only and without  fee is  hereby
granted, provided that this copyright notice and the original authors'
names  appear  on  all  copies  and   supporting   documentation.   If
individual  files  are  separated  from  this  distribution  directory
structure, this copyright notice must be included.  For any other uses
of  this software, in  original or  modified form, including  but  not
limited to distribution in whole or in part, specific prior permission
must be  obtained  from  MIT.   These  programs  shall  not  be  used,
rewritten,  or  adapted  as  the  basis  of a  commercial software  or
hardware  product without  first  obtaining  appropriate licenses from
MIT.  MIT.  makes no  representations about  the  suitability of  this
software for any purpose.  It is  provided "as is" without  express or
implied warranty.

******************************************************************/


/*
*	l u d . c			dead@media-lab		87/02/27
*
*		LU Decomposition:  Decompose, Solve and Improve
*/

#ifndef lint
static char copyright[] =
"Copyright (c) 1987 by David T. Chen.  All rights reserved.";
#endif

#include <stdio.h>
#include <math.h>
#include "matrix.h"

#define EPSILON		(1.0e-16)


/*
==========================================================================
LUdecomposition() -- b = Ax

	Given A and b, x is found.
	The function returns the singularity status of A. 

==========================================================================
*/
LUdecomposition( b, A, x )
VectorType *b;
MatrixType *A;
VectorType *x;
{
IndexType IX;
MatrixType LU;

	MatrixType_Copy( &LU, A );

	if (Decompose( &LU, &IX ) == SINGULAR)
		return( SINGULAR );

	if (Solve( b, &LU, &IX, x ) == SINGULAR)
		return( SINGULAR );

	/*
	*	Not neccessary if using doubles
	*
	* 	if (Improve( b, A, &LU, &IX, 0, x ) == SINGULAR)
	*		return( SINGULAR );
	*/

	return( NOT_SINGULAR );
}
/* LUdecomposition */



/*
==========================================================================
Decompose() -- b = Ax

	A is returned in the compact LU decomposition form, L-I and U.

	Because `partial pivoting' is used, the row order
	of LU is returned in IX.

	The function returns the singularity status of A. 

==========================================================================
*/
Decompose( A_matrix, IX_index )
MatrixType *A_matrix;
IndexType *IX_index;
{
register i, j;
int k;
register index_i;
int index_k;
int MS = MatrixType_Get_N(A_matrix);
register MatrixData A = MatrixType_Get_Data(A_matrix);
register IndexData IX = IndexType_Get_Data(IX_index);
RobotType tmp;

	if (MS != MatrixType_Get_M(A_matrix))
	{	fprintf(stderr, "Decompose: matrix `A' not square\n");
		return( SINGULAR );
	}
	IndexType_Set_N(IX_index, MS);

	/* Initialize Index Array */
	for (i=0; i<MS; i++) IX[i] = i;

	k = 0;
	do
	{	if (PivotRowSearch(MS, A, IX, k) == SINGULAR)
			return( SINGULAR );
	
		index_k = IX[k];
		A[ index_k ][k] = 1.0 / A[ index_k ][k];

		for (i=k+1; i<MS; i++)
		{
			index_i = IX[i];
			A[ index_i ][k] *= A[ index_k ][k];

			tmp = A[ index_i ][k];

			for (j=k+1; j<MS; j++)
				A[ index_i ][j] -= tmp * A[ index_k ][j];
		}
		k++;

	} while (k < MS);

	return( NOT_SINGULAR );
}
/* Decompose */



/*
==========================================================================
Solve() -- b = Ax

	Uses the LU factorization from Decompose() to find an approximate
	solution for 'x' given 'b'.

==========================================================================
*/
Solve( b_vector, LU_matrix, IX_index, x_vector )
VectorType *b_vector;
MatrixType *LU_matrix;
IndexType *IX_index;
VectorType *x_vector;
{
register i, j;
register index_i;
int MS = MatrixType_Get_N(LU_matrix);
register MatrixData LU = MatrixType_Get_Data(LU_matrix);
VectorData b = VectorType_Get_Data(b_vector);
register VectorData x = VectorType_Get_Data(x_vector);
IndexData IX = IndexType_Get_Data(IX_index);
VectorType y_vector;
register VectorData y = VectorType_Get_Data(&y_vector);

	if (MS != MatrixType_Get_M(LU_matrix))
	{	fprintf(stderr, "Solve: matrix `LU' not square\n");
		return( SINGULAR );
	}
	if (MS != VectorType_Get_N(b_vector))
	{	fprintf(stderr, "Solve: vector `b' improper rank\n");
		return( SINGULAR );
	}
	VectorType_Set_N(x_vector, MS);
	VectorType_Set_N(&y_vector, MS);

	/* Forward Elimination:  Ly = b */
	for (i=0; i<MS; i++)
	{
		index_i = IX[i];

		y[i] = b[index_i];
		for (j=0; j<i; j++)
			y[i] -= LU[index_i][j] * y[j];
	}

	/* Back Substitution:  Ux = y */
	for (i=MS-1; i>=0; i--)
	{
		index_i = IX[i];

		x[i] = y[i];
		for (j=i+1; j<MS; j++)
			x[i] -= LU[index_i][j] * x[j];

		x[i] *= LU[index_i][i];
	}

	return( NOT_SINGULAR );
}
/* Solve */



/*
==========================================================================
Improve() -- b = Ax

	Uses A, b, and the LU factorization from Decompose() to iteratively
	improve the approximate solution x returned by Solve().

==========================================================================
*/
Improve( b_vector, A_matrix, LU, IX, max_times, x_vector )

VectorType *b_vector;
MatrixType *A_matrix;
MatrixType *LU;
IndexType *IX;
int max_times;
VectorType *x_vector;
{
int i;
register j, k;
int MS = MatrixType_Get_N(A_matrix);
VectorData b = VectorType_Get_Data(b_vector);
register MatrixData A = MatrixType_Get_Data(A_matrix);
register VectorData x = VectorType_Get_Data(x_vector);
VectorType r_vector;
VectorData r = VectorType_Get_Data(&r_vector);
VectorType dx_vector;
register VectorData dx = VectorType_Get_Data(&dx_vector);
RobotType norm_x;
RobotType norm_dx;
RobotType tmp;
double d_tmp;

	if (MS != MatrixType_Get_M(A_matrix))
	{	fprintf(stderr, "Improve: matrix `A' not square\n");
		return( SINGULAR );
	}
	if (MS != MatrixType_Get_N(LU))
	{	fprintf(stderr, "Improve: matrix `LU' improper n-rank\n");
		return( SINGULAR );
	}
	if (MS != MatrixType_Get_M(LU))
	{	fprintf(stderr, "Improve: matrix `LU' improper m-rank\n");
		return( SINGULAR );
	}
	if (MS != VectorType_Get_N(b_vector))
	{	fprintf(stderr, "Improve: vector `b' improper rank\n");
		return( SINGULAR );
	}
	if (MS != VectorType_Get_N(x_vector))
	{	fprintf(stderr, "Improve: vector `x' improper rank\n");
		return( SINGULAR );
	}
	if (MS != IndexType_Get_N(IX))
	{	fprintf(stderr, "Improve: index `IX' improper rank\n");
		return( SINGULAR );
	}
	VectorType_Set_N(&r_vector, MS);
	VectorType_Set_N(&dx_vector, MS);

	if (max_times >= MAX_IMPROVE) max_times = MAX_IMPROVE;

	/* Find norm */
	norm_x = 0.0;
	for (i=0; i<MS; i++)
		if (norm_x < abs_robot(x[i])) norm_x = abs_robot(x[i]);

	if (norm_x == 0.0)
		return( NOT_SINGULAR );

	for (i=0; i<max_times; i++)
	{
		/* Find residual, r = b - Ax */
		for (j=0; j<MS; j++)
		{
			d_tmp = 0.0;
			for (k=0; k<MS; k++)
				d_tmp += A[j][k] * x[k];

			d_tmp = b[j] - d_tmp;
			r[j] = d_tmp;
		}

		Solve( &r_vector, LU, IX, &dx_vector );

		norm_dx = 0.0;
		for (j=0; j<MS; j++)
		{
			tmp = x[j];
			x[j] += dx[j];
			if (norm_dx < abs_robot(x[j] - tmp))
				norm_dx = abs_robot(x[j] - tmp);
		}

		if (norm_dx < EPSILON * norm_x)
			return( NOT_SINGULAR );
	}

	if (max_times == MAX_IMPROVE)
		fprintf(stderr, "Improve:  WARNING, solution did not converge.\n");

	return( NOT_SINGULAR );
}
/* Improve */




/*
==========================================================================
PivotRowSearch( )

==========================================================================
*/
static PivotRowSearch( MS, A, IX, row_counter )
int MS;
register MatrixData A;
register IndexData IX;
int row_counter;
{
int i;					/* current row search */
int lr = row_counter;	/* largest row number */
int tmp;
RobotType largest_element = 0.0;
RobotType compare_element;

	/* Search */
	for (i=row_counter; i<MS; i++)
	{
		compare_element = abs_robot(A[ IX[i] ][ row_counter ]);

		if (largest_element < compare_element)
		{	lr = i;
			largest_element = compare_element;
		}
	}

	/* Swap */
	if (lr != row_counter)
	{
		tmp = IX[ lr ];
		IX[ lr ] = IX[ row_counter ];
		IX[ row_counter ] = tmp;
	}

	/* Check for Singularity */
	if (abs_robot(A[ IX[row_counter] ][ row_counter ]) < EPSILON)
	{	return( SINGULAR );
	}
	else return( NOT_SINGULAR );
}
/* PivotRowSearch */



/*
==========================================================================
Invert( ) -- Invert a matrix.

==========================================================================
*/
Invert( A, Ainv_matrix )

MatrixType *A;
MatrixType *Ainv_matrix;
{
register i, j;
MatrixType LU;
IndexType IX;
int n = MatrixType_Get_N(A);
register MatrixData Ainv = MatrixType_Get_Data(Ainv_matrix);
VectorType b_vector;
register VectorData b = VectorType_Get_Data(&b_vector);
VectorType x_vector;
register VectorData x = VectorType_Get_Data(&x_vector);

	if (n != MatrixType_Get_M(A))
	{	fprintf(stderr, "Invert: matrix not square.\n");
		return( SINGULAR );
	}
	MatrixType_Set_N_and_M(Ainv_matrix, n, n);
	VectorType_Set_N(&b_vector, n);

	MatrixType_Copy( &LU, A );

	if (Decompose( &LU, &IX ) == SINGULAR)
		return( SINGULAR );

	for (i=0; i<n; i++)
	{
		for (j=0; j<n; j++)
		{
			if (i == j)
			{	b[j] = 1.0;
			}
			else b[j] = 0.0;
		}

		if (Solve( &b_vector, &LU, &IX, &x_vector ) == SINGULAR)
			return( SINGULAR );

		for (j=0; j<n; j++)
			Ainv[j][i] = x[j];
	}
	return( NOT_SINGULAR );
}
/* Invert */





